project = "cabinetry"
author = "cabinetry developers"
copyright = f"2020, {author}"
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx_rtd_theme",
    "sphinx_copybutton",
    "sphinx-jsonschema",
    "sphinx_click",
]

import sphinx_rtd_theme

html_theme = "sphinx_rtd_theme"
